// TODO: Fix threading - Don't keep a suspended thread around!
// TODO: Clean up XML Comments.
// TODO: Create unit tests (Cache pages before overwriting!)

{ *******************************************************************************
  Copyright (c) MicroWorks Inc. 2003 - 2010
  Free for use and modification with Smart Dongle hardware

  mwi_sd.pas

  Source for TmwiSmartDongle and related components.
  ******************************************************************************** }

unit mwi_sd;

interface

uses
  classes,
  windows,
  sysutils,
  usk,
  Winsock,
  Forms,
  StrUtils,
  sutils,
  EncdDecd;

const
  EEPROM_PAGE_SIZE = 16;

  SD_ERR_LOAD_LIB = 1000;
  SD_ERR_INVALID_LIB = 1001;
  SD_ERR_INVALID_KEYS = 1002;
  SD_ERR_INVALID_INTERVAL = 1003;
  SD_ERR_INVALID_KEY_STR = 1004;
  SD_ERR_REMOTE_NOT_PRESENT = 1005;
  SD_ERR_NOT_AVAILABLE_FOR_REMOTE = 1006;
  SD_ERR_REMOTE_CONNECTION = 1007;

  REQ_DONGLE_CHECK           = '1000';
  RESP_DONGLE_NOT_PRESENT    = '1001';
  RESP_DONGLE_PRESENT        = '1002';
  REQ_READ_STRING            = '1003';
  REQ_WRITE_STRING           = '1004';
  REQ_READ_INT               = '1005';
  REQ_WRITE_INT              = '1006';
  REQ_READ_DBL               = '1007';
  REQ_WRITE_DBL              = '1008';
  REQ_READ                   = '1009';
  REQ_WRITE_PAGE             = '1010';
  RESP_ERROR                 = '1011';

type

  TsdThread = class;

/// <summary>
///  A Smart Dongle related error
/// </summary>
  ESDError = class(Exception)
  public
    ErrorCode: integer;
    constructor Create(const Msg: string; iErrorCode: integer);
  end;

///  The three states of the LED on the Smart Dongle.
 TsdLedColor = (sdNone, sdGreen, sdRed);

  TsdOnThreadError = procedure(ErrorCode: integer; ErrorMsg: string) of Object;

  TsdOnThreadSuccess = procedure() of Object;

  TUInt64 = record
    buffer: array [0 .. 7] of byte;
  end;

///  For working with a local USB Smart Dongle.
  TmwiSmartDongle = class(TComponent)
  private
    FpThread: TsdThread;
    FhLib: cardinal;
    FpRead: TuskRead;
    FpWrite: TuskWrite;
    FpGetVersion: TuskGetVersion;
    FpLEDRed: TuskLEDRed;
    FpLEDGreen: TuskLEDGreen;
    FpLEDOff: TuskLEDOff;
    FpEepromSize: TuskEepromSize;
    FstrKey1: string;
    FstrKey2: string;
    FabtKey1: array [0 .. 7] of byte;
    FabtKey2: array [0 .. 7] of byte;
    FpOnThreadError: TsdOnThreadError;
    FpOnThreadSuccess: TsdOnThreadSuccess;
    FbThreadActive: boolean;
    FiInterval: Cardinal;

    procedure Load(lib: string);
    procedure SetKey1(const Value: string);
    procedure SetKey2(const Value: string);
    procedure CheckErr(iStatus: integer);
    procedure SetThreadActive(const Value: boolean);
    /// Loads the Smart Dongle driver if it is not loaded yet.
    procedure CheckLoaded;
    function GetVersion: _USKDllVersionInfo;
    procedure Unload;

  protected
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure Authenticate();
    /// <summary>Retrieves the user memory area size (EEPROM) from the dongle.</summary>
    function GetEepromSize: integer;

    procedure SetLEDColor(Color: TsdLedColor);
    procedure LEDRed;
    procedure LEDGreen;
    procedure LEDOff;

    procedure DoOnThreadError(ErrorCode: integer; ErrorMsg: string);

    procedure DoOnThreadSuccess();

    procedure Read(Offset: dword; buffer: Pointer; BufferSize: integer);

    /// <summary>
    /// Writes 1 or multiple pages to EEPROM.
    /// </summary>
    procedure Write(Offset: dword; buffer: Pointer; BufferSize: integer);

    function ReadString(Page: integer): string;

    function ReadAnsiString(Page: integer): AnsiString;

    function ReadWideString(Page: integer): WideString;

    function ReadInteger(Page: integer): integer;

    function ReadDouble(Page: integer): double;

    procedure WriteInteger(Page, Value: integer);

    procedure WriteDouble(Page: integer; Value: double);

    procedure WriteString(Page: integer; Value: String);
    procedure WriteAnsiString(Page: integer; Value: AnsiString);
    procedure WriteWideString(Page: integer; Value: WideString);


    /// Retrieves the user memory area size (EEPROM) from the dongle.
    property EepromSize: integer read GetEepromSize;

    property Version: _USKDllVersionInfo read GetVersion;

  published
    property Key1: string read FstrKey1 write SetKey1;
    property Key2: string read FstrKey2 write SetKey2;
    property ThreadInterval: Cardinal read FiInterval write FiInterval;
    property ThreadActive: Boolean read FbThreadActive write SetThreadActive;

    // Events
    property OnThreadError: TsdOnThreadError read FpOnThreadError
      write FpOnThreadError;
    property OnThreadSuccess: TsdOnThreadSuccess read FpOnThreadSuccess
      write FpOnThreadSuccess;
  end;

  TsdThread = class(TThread)
  private
    FpOwner: TmwiSmartDongle;
  protected
    constructor Create(Owner: TmwiSmartDongle);
    procedure Execute; override;
  end;

function GetErrDesc(iStatus: integer): string;

function StrToUInt64(strValue: string): TUInt64;

procedure Register();

implementation

procedure Register;
begin
  RegisterComponents('MicroWorks', [TmwiSmartDongle]);
end;

{ TmwiSmartDongle }
constructor TmwiSmartDongle.Create(AOwner: TComponent);
begin
  inherited;

  FbThreadActive := false;
end;

destructor TmwiSmartDongle.Destroy;
begin
  Unload;

  if (FpThread <> nil) then
    try
      FpThread.Terminate();
      FreeAndNil(FpThread);
    except
    end;

  inherited;
end;

procedure TmwiSmartDongle.CheckLoaded();
begin
  // Make sure the library is loaded.
  if (FhLib = 0) then
    Load('usk.dll');
end;

/// <summary>
/// This method reads from the first EEPROM page using the keys
/// provided in Key1 and Key2 properties. If not successful an exception
/// will occur.
/// </summary>
procedure TmwiSmartDongle.Authenticate();
var
  buffer: PByte;
  size: integer;
begin
  buffer := nil;

  try
    // Allocate enough mem for 1 page.
    GetMem(buffer, EEPROM_PAGE_SIZE);
    ZeroMemory(buffer, EEPROM_PAGE_SIZE);

    CheckLoaded();

    // verify keys have been set.
    if (FstrKey1 = '') or (FstrKey2 = '') then
      raise ESDError.Create
        ('Key values must be set prior to calling Authenticate',
        SD_ERR_INVALID_KEYS);

    // Read the first page ( offset 0 ).
    // for testing temp := Int64( FabtKey1 );
    size := EEPROM_PAGE_SIZE;
    CheckErr(FpRead(Int64(FabtKey1), Int64(FabtKey2), 0, buffer, size));
  finally
    if (buffer <> nil) then
      FreeMem(buffer);
  end;
end; // Authenticate

procedure TmwiSmartDongle.LEDGreen;
begin
  SetLEDColor(sdGreen);
end;

procedure TmwiSmartDongle.LEDOff;
begin
  SetLEDColor(sdNone);
end;

procedure TmwiSmartDongle.LEDRed;
begin
  SetLEDColor(sdRed);
end;

/// <summary>
/// Dynamically loads the usk.dll library.
/// </summary>
procedure TmwiSmartDongle.Load(lib: string);
begin
  inherited;

  Unload();

  // Load library dynamically rather than statically so
  // that we leave open the option of customizing the location of the
  // USK.DLL as well as custom error handling.
  FhLib := LoadLibrary(PChar(lib));

  // If lib is loaded, then get function addresses.
  if (FhLib <> 0) then
  begin

    // Read
    @FpRead := GetProcAddress(FhLib, 'Read');

    // Write
    @FpWrite := GetProcAddress(FhLib, 'Write');

    // GetVersion
    @FpGetVersion := GetProcAddress(FhLib, 'DllGetVersion');

    // LED Green
    @FpLEDGreen := GetProcAddress(FhLib, 'LedGreen');

    // LED Red
    @FpLEDRed := GetProcAddress(FhLib, 'LedRed');

    // LED Off
    @FpLEDOff := GetProcAddress(FhLib, 'LedOff');

    // Eeprom Size
    @FpEepromSize := GetProcAddress(FhLib, 'GetEpromSize');

    if  ( @FpEepromSize = nil ) or
        ( @FpLEDOff = nil ) or
        ( @FpLEDRed = nil ) or
        ( @FpLEDGreen = nil ) or
        ( @FpGetVersion = nil ) or
        ( @FpRead = nil ) or
        ( @FpWrite = nil ) then
      raise ESDError.Create('Smart Dongle Error: Invalid Library.',
        SD_ERR_INVALID_LIB);
  end // if( FhLib <> 0 ) then

  else
    raise ESDError.Create('Smart Dongle Error: Unable to load library.',
      SD_ERR_LOAD_LIB);
end; // Load

/// <summary>
/// Set method for Key1 property, converts string to array of byte
/// and stores.
/// </summary>
procedure TmwiSmartDongle.SetKey1(const Value: string);
var
  Temp: TUInt64;
begin
  ZeroMemory(@Temp, sizeof(TUInt64));

  Temp := StrToUInt64(Value);

  CopyMemory(@FabtKey1, @(Temp.buffer), 8);

  FstrKey1 := Value;
end;

/// <summary>
/// Set method for Key2 property, converts string to array of byte
/// and stores.
/// </summary>
procedure TmwiSmartDongle.SetKey2(const Value: string);
var
  Temp: TUInt64;
begin
  ZeroMemory(@Temp, sizeof(TUInt64));

  Temp := StrToUInt64(Value);

  CopyMemory(@FabtKey2, @(Temp.buffer), 8);

  FstrKey2 := Value;
end;

/// <summary>
/// Changes the dongle LED color.
/// </summary>
procedure TmwiSmartDongle.SetLEDColor(Color: TsdLedColor);
begin

  // Make sure the library is loaded.
  CheckLoaded();

  case Color of
    sdGreen:
      CheckErr(FpLEDGreen());
    sdRed:
      CheckErr(FpLEDRed());
    sdNone:
      CheckErr(FpLEDOff());
  end; // case
end; // SetLEDColor

/// <summary>
/// Utility method for checking function call status result.
/// </summary>
procedure TmwiSmartDongle.CheckErr(iStatus: integer);
begin
  if (iStatus <> USK_OK) then
    raise ESDError.Create('Smart Dongle Error. ' + GetErrDesc(iStatus),
      iStatus);
end;

/// <summary>
/// Utility function to retrieve error message.
/// </summary>
function GetErrDesc(iStatus: integer): string;
begin
  case iStatus of
    SD_ERR_LOAD_LIB:
      result := 'Unable to load the USK.DLL library.';

    SD_ERR_INVALID_LIB:
      result := 'Invalid USK.DLL library.';

    SD_ERR_INVALID_KEYS:
      result := 'Key1 and Key2 properties must be set.';

    SD_ERR_INVALID_INTERVAL:
      result := 'The thread interval must be 1 or greater.';

    SD_ERR_INVALID_KEY_STR:
      result := 'StrToUInt64() invalid string.';

    SD_ERR_REMOTE_NOT_PRESENT:
      result := 'Dongle not present';

    SD_ERR_NOT_AVAILABLE_FOR_REMOTE:
      result := 'The request is not supported for remote sessions.';

    SD_ERR_REMOTE_CONNECTION:
      result := 'A connection cannot be made to the Smart Dongle server.';

    USK_ERR_OPEN:
      result := 'Can''t open dongle.';

    USK_ERR_IS_OPEN:
      result := 'Dongle is already open.';

    USK_ERR_MEM:
      result := 'Memory allocation error.';

    USK_ERR_EEPROM_SIZE:
      result := 'Address offset or buffer size exceeds EEPROM memory.';

    USK_ERR_SECURITY_STATE_TIMEOUT:
      result := 'Timeout waiting for state change.';

    USK_ERR_KEY_STATE_TIMEOUT:
      result := 'Timeout waiting for state change in QTS-195 and Host Challenge handshake.';

    USK_ERR_TWC_RESET_TIMEOUT:
      result := 'Timeout waiting for TWC to reset';

    USK_ERR_BUSY_TIMEOUT:
      result := 'Timeout waiting for task_loop to be flagged as inactive.';

    USK_ERR_NOT_ENABLED:
      result := 'The Dongle is Not in ENABLED Mode.';

    USK_ERR_EEPROM_READ_TIMEOUT:
      result := 'Timeout Waiting For Read Flag(s) to be set.';

    USK_ERR_EEPROM_READ:
      result := 'An EEPROM Read ERROR Occurred.';

    USK_ERR_EEPROM_WRITE:
      result := 'An EEPROM Write ERROR Occurred.';

    USK_ERR_USK_VALIDATE:
      result := 'Dongle failed to validate against user keys.';

    USK_ERR_HOST_VALIDATE:
      result := 'Host failed to validate against dongle keys';

    USK_ERR_DISABLE:
      result := 'An error occurred when attempting to disable QTS-195.';

    USK_ERR_INVALID_KEYS:
      result := 'Invalid keys';
  end; // case
end; // GetErrDesc

/// <summary>
/// Set method for ThreadActive property. When true, background thread
/// is started which checks dongle status.
/// </summary>
procedure TmwiSmartDongle.SetThreadActive(const Value: boolean);
begin
  if (Value <> FbThreadActive) then
  begin
    FbThreadActive := Value;
    if Assigned(FpThread) then
      FreeAndNil(FpThread);

    if (Value) then
    begin
      if (FiInterval < 1) then
        raise ESDError.Create('Smart Dongle Error. ' +
          GetErrDesc(SD_ERR_INVALID_INTERVAL), SD_ERR_INVALID_INTERVAL);

      FpThread := TsdThread.Create(self);
    end
  end;
end;

/// <summary>
/// OnThreadError event. Called by background thread when an error occurs.
/// </summary>
procedure TmwiSmartDongle.DoOnThreadError(ErrorCode: integer; ErrorMsg: string);
begin
  if (Assigned(FpOnThreadError)) then
    FpOnThreadError(ErrorCode, ErrorMsg);
end;

/// <summary>
/// OnThreadSuccess event. Called by background thread when it successfully
/// validates the dongle.
/// </summary>
procedure TmwiSmartDongle.DoOnThreadSuccess();
begin
  if (Assigned(FpOnThreadSuccess)) then
    FpOnThreadSuccess();
end;

/// <summary>Retrieves the user memory area size (EEPROM) from the dongle.</summary>
function TmwiSmartDongle.GetEepromSize: integer;
begin
  CheckLoaded();

  result := FpEepromSize(Int64(FabtKey1), Int64(FabtKey2));
end;

/// <summary>
/// Reads 1 or multiple pages from EEPROM.
/// </summary>
procedure TmwiSmartDongle.Read(Offset: dword; buffer: Pointer;
  BufferSize: integer);
begin
  CheckLoaded();

  // verify keys have been set.
  if (FstrKey1 = '') or (FstrKey2 = '') then
    raise ESDError.Create
      ('Key values must be set prior to calling Authenticate',
      SD_ERR_INVALID_KEYS);

  // Read from the user EEPROM space.
  CheckErr(FpRead(Int64(FabtKey1), Int64(FabtKey2), Offset, buffer,
    BufferSize));
end;

procedure TmwiSmartDongle.Write(Offset: dword; buffer: Pointer;
  BufferSize: integer);
begin
  CheckLoaded();

  // verify keys have been set.
  if (FstrKey1 = '') or (FstrKey2 = '') then
    raise ESDError.Create
      ('Key values must be set prior to calling Authenticate',
      SD_ERR_INVALID_KEYS);

  // Write to the user EEPROM space.
  CheckErr(FpWrite(Int64(FabtKey1), Int64(FabtKey2), Offset, buffer,
    BufferSize));
end;

/// <summary>
/// Retrieves the version info for the USB driver.
/// </summary>
function TmwiSmartDongle.GetVersion: _USKDllVersionInfo;
begin
  CheckLoaded();

  FpGetVersion(result);
end;

function TmwiSmartDongle.ReadWideString(Page: integer): WideString;
var
  buffer: array [0 .. EEPROM_PAGE_SIZE - 1] of byte;
begin
  ZeroMemory(@buffer, EEPROM_PAGE_SIZE);

  Read(Page * EEPROM_PAGE_SIZE, @buffer, EEPROM_PAGE_SIZE);

  result := StrPas(PWideChar(@buffer));
end;

function TmwiSmartDongle.ReadAnsiString(Page: integer): AnsiString;
var
  buffer: array [0 .. EEPROM_PAGE_SIZE - 1] of byte;
begin
  ZeroMemory(@buffer, EEPROM_PAGE_SIZE);

  Read(Page * EEPROM_PAGE_SIZE, @buffer, EEPROM_PAGE_SIZE);

  result := StrPas(PAnsiChar(@buffer));
end;


function TmwiSmartDongle.ReadString(Page: integer): string;
var
  buffer: array [0 .. EEPROM_PAGE_SIZE - 1] of byte;
begin
  ZeroMemory(@buffer, EEPROM_PAGE_SIZE);

  Read(Page * EEPROM_PAGE_SIZE, @buffer, EEPROM_PAGE_SIZE);

  result := StrPas(PChar(@buffer));
end;

procedure TmwiSmartDongle.Unload;
begin
  if (FhLib <> 0) then
    FreeLibrary(FhLib);
end;

procedure TmwiSmartDongle.WriteWideString(Page: integer; Value: WideString);
begin
  if Length(Value) * SizeOf(WideChar) <= EEPROM_PAGE_SIZE then
    Write(Page * EEPROM_PAGE_SIZE, PWideChar(Value), EEPROM_PAGE_SIZE)
  else
    raise Exception.Create('Size of Value is greater than EEPROM_PAGE_SIZE.');
end;

procedure TmwiSmartDongle.WriteAnsiString(Page: integer; Value: AnsiString);
begin
  if Length(Value) * SizeOf(AnsiChar) <= EEPROM_PAGE_SIZE then
    Write(Page * EEPROM_PAGE_SIZE, PAnsiChar(Value), EEPROM_PAGE_SIZE)
  else
    raise Exception.Create('Size of Value is greater than EEPROM_PAGE_SIZE.');
end;

procedure TmwiSmartDongle.WriteString(Page: integer; Value: String);
begin
  if Length(Value) * SizeOf(Char) <= EEPROM_PAGE_SIZE then
    Write(Page * EEPROM_PAGE_SIZE, PChar(Value), EEPROM_PAGE_SIZE)
  else
    raise Exception.Create('Size of Value is greater than EEPROM_PAGE_SIZE.');
end;

function TmwiSmartDongle.ReadInteger(Page: integer): integer;
var
  buffer: array [0 .. EEPROM_PAGE_SIZE - 1] of byte;
begin
  ZeroMemory(@buffer, EEPROM_PAGE_SIZE);

  Read(Page * EEPROM_PAGE_SIZE, @buffer, EEPROM_PAGE_SIZE);

  CopyMemory(@result, @buffer, sizeof(result));
end;

procedure TmwiSmartDongle.WriteInteger(Page: integer; Value: integer);
var
  buffer: array [0 .. EEPROM_PAGE_SIZE - 1] of byte;
begin
  ZeroMemory(@buffer, EEPROM_PAGE_SIZE);

  CopyMemory(@buffer, @Value, sizeof(Value));

  Write(Page * EEPROM_PAGE_SIZE, @buffer, EEPROM_PAGE_SIZE);
end;

function TmwiSmartDongle.ReadDouble(Page: integer): double;
var
  buffer: array [0 .. EEPROM_PAGE_SIZE - 1] of byte;
begin
  ZeroMemory(@buffer, EEPROM_PAGE_SIZE);

  Read(Page * EEPROM_PAGE_SIZE, @buffer, EEPROM_PAGE_SIZE);

  CopyMemory(@result, @buffer, sizeof(result));
end;

procedure TmwiSmartDongle.WriteDouble(Page: integer; Value: double);
var
  buffer: array [0 .. EEPROM_PAGE_SIZE - 1] of byte;
begin
  ZeroMemory(@buffer, EEPROM_PAGE_SIZE);

  CopyMemory(@buffer, @Value, sizeof(Value));

  Write(Page * EEPROM_PAGE_SIZE, @buffer, EEPROM_PAGE_SIZE);
end;

{ ESDError }

constructor ESDError.Create(const Msg: string; iErrorCode: integer);
begin
  self.Message := Msg;
  ErrorCode := iErrorCode;
end;

{ TsdThread }

constructor TsdThread.Create(Owner: TmwiSmartDongle);
begin
  FpOwner := Owner;
  inherited Create(false);
end;

/// <summary>
/// Thread body, checks for dongle presence on timed interval.
/// </summary>
procedure TsdThread.Execute;
begin
  while (not Terminated) and (FpOwner <> nil) do
  begin

    // wait for interval, convert to msec.
    Sleep(FpOwner.ThreadInterval * 1000);

    try
      FpOwner.Authenticate();
      FpOwner.DoOnThreadSuccess();
    except
      on e: ESDError do
      begin
        FpOwner.DoOnThreadError(e.ErrorCode, e.Message);
      end;
    end;
  end; // while
end;

/// <summary>
/// Utility function converts a hex string to UINT64 buffer. strValue
/// must be a string in the form 0x0000000000000000
/// </summary>
function StrToUInt64(strValue: string): TUInt64;
var
  strByte: string;
  I: integer;
  btTemp: byte;
begin

  // change all to lower case for easier parsing.
  strValue := LowerCase(strValue);

  // remove the leading 0x.
  Delete(strValue, 1, 2);

  // verify string is 16 chars
  if (Length(strValue) <> 16) then
    raise ESDError.Create('Smart Dongle Error. ' +
      GetErrDesc(SD_ERR_INVALID_KEY_STR), SD_ERR_INVALID_KEY_STR);

  for I := 8 downto 1 do
  begin
    strByte := strValue[(I * 2) - 1] + strValue[I * 2];
    HexToBin(PChar(strByte), @btTemp, 1);

    result.buffer[8 - I] := btTemp;
  end; // for
end; // StrToUInt64

end.
